"""thaiaddress: A python parser for Thai address"""

# PEP0440 compatible formatted version, see:
# https://www.python.org/dev/peps/pep-0440/
#
# Generic release markers:
#   X.Y
#   X.Y.Z   # For bugfix releases
#
# Admissible pre-release markers:
#   X.YaN   # Alpha release
#   X.YbN   # Beta release
#   X.YrcN  # Release Candidate
#   X.Y     # Final release
#
# Dev branch marker is: 'X.Y.devN' where N is an integer.
#

__version__ = "0.1.3"

from .parser import parse
from .utils import (
    preprocess,
    is_stopword,
    merge_tokens,
    merge_labels,
)
from .train import (
    train,
)



import joblib
import deepcut
import jsonlines

from sklearn_crfsuite import metrics, CRF
from sklearn.model_selection import train_test_split

from .parser import tokens_to_features
from .utils import range_intersect


LABELS_MAP = {
    "ชื่อ": "NAME",
    "ที่อยู่ย่อย": "ADDR",
    "ที่อยู่ - พื้นที่": "LOC",
    "รหัสไปรษณีย์": "POST",
    "เบอร์โทร": "PHONE",
    "อีเมล์": "EMAIL",
}
LABELS = list(LABELS_MAP.values())


def address_to_token(address: dict):
    """
    Transform address dictionary to a list of tokens

    Input
    -----
    >>> address = {
        "text": ...,
        "labels": [[start1, stop1, label1], [start2, stop2, label2]]
    }

    Output
    ------
    >>> [(token1, label1), (token2, label2), ...]
    """
    if address["labels"] != []:
        tokens = []
        s = 0
        for token in deepcut.tokenize(address["text"]):
            start = s
            stop = s + len(token)

            label = "O"
            for s, st, c in address["labels"]:
                if range_intersect(range(start, stop), range(s, st)):
                    label = c
            tokens.append((token, label))
            s = stop
        return tokens
    else:
        return None


def address_to_feature(address: dict):
    """
    Transform address dictionary to features and labels
    """
    tokens = address_to_token(address)
    features = [tokens_to_features(tokens, i) for i in range(len(tokens))]
    labels = [LABELS_MAP.get(label, "O") for _, label in tokens]
    return features, labels


def addresses_to_features(addresses: list):
    """
    Transform list of addresses to features and labels
    """
    X, y = [], []
    for address in addresses:
        # check if already labeled
        if len(address["labels"]) > 0:
            features, labels = address_to_feature(address)
            X.append(features)
            y.append(labels)
    return X, y


def train(file_path: str, model_path: str = None):
    """
    Training CRF model from a given ``file_path``
    """
    addresses = []
    with jsonlines.open(file_path) as reader:
        for obj in reader:
            addresses.append(obj)
    addresses_train, addresses_val = train_test_split(
        addresses, test_size=0.25, random_state=42
    )

    X_train, y_train = addresses_to_features(addresses_train)
    X_val, y_val = addresses_to_features(addresses_val)

    crf = CRF(c1=0.2, c2=0.2, max_iterations=100, all_possible_transitions=True)
    crf.fit(X_train, y_train)

    # prediction score on validation set
    y_pred = crf.predict(X_val)
    f1_score = metrics.flat_f1_score(
        y_val, y_pred, average="weighted", labels=[l for l in LABELS if l != "O"]
    )
    print("Flat F1-Score on validation set = {}".format(f1_score))

    if model_path:
        joblib.dump(crf, model_path)
        print("Save model to {}".format(model_path))

    return crf



from itertools import groupby
import numpy as np
import deepcut
from pythainlp.util import isthai
from pythainlp.corpus import thai_stopwords


def preprocess(text: str) -> str:
    """
    Generalized function to preprocess an input
    """
    text = text.strip()
    text = text.replace("จัดส่ง", "")
    text = text.replace("ชือ.", "")
    text = text.replace("ชื่อ ", "")
    text = text.replace("ผู้รับ", "")
    text = text.replace("ส่งที่ ", " ")
    text = text.replace("ที่อยู่ ", " ")
    text = text.replace("ที้อยุ่ ", " ")
    text = text.replace("ส่งของที่ ", " ")
    text = text.replace("ส่งมาที่", " ")
    text = text.replace("\n-", " ")
    text = text.replace("\n", " ")
    text = text.replace(": ", " ")
    text = text.replace("(", "")
    text = text.replace(")", "")
    text = " ".join([t for t in text.strip().split(" ") if t.strip() != ""])
    return text


def clean_location_text(text: str) -> str:
    """
    Clean location before using fuzzy string match
    """
    text = text.replace("แขวง", " ")
    text = text.replace("เขต", " ")
    text = text.replace("อำเภอ", " ")
    text = text.replace("ตำบล", " ")
    text = text.replace("ต.", " ")
    text = text.replace("ตฺ", "ต.")
    text = text.replace("อ.", " ")
    text = text.replace("จ.", " ")
    text = text.replace("คอหงส์", "คอหงษ์")
    text = text.replace("กทม.", "กรุงเทพมหานคร")
    text = text.replace("กทม", "กรุงเทพมหานคร")
    return text


def get_digit(text: str) -> str:
    """
    Get digit output from a given text
    """
    return "".join([c for c in text if c.isdigit()])


def is_stopword(word: str) -> bool:  # เช็คว่าเป็นคำฟุ่มเฟือย
    """
    Check if a word is stop word or not using PyThaiNLP

    Reference
    ----------
    Pythainlp, https://github.com/PyThaiNLP/pythainlp
    """
    return word in thai_stopwords()


def range_intersect(r1: range, r2: range):
    """
    Check if range is intersected

    References
    ----------
    Stack Overflow, https://stackoverflow.com/questions/6821156/how-to-find-range-overlap-in-python
    """
    return range(max(r1.start, r2.start), min(r1.stop, r2.stop)) or None


def merge_labels(preds: list):
    """
    Get merged labels and merge tuple to merge tokens
    """
    preds = list(np.ravel(preds))
    merge, labels = [], []
    s = 0
    for label, g in groupby(preds):
        g = list(g)
        labels.append(label)
        if len(g) > 1:
            merge.append((s, s + len(g)))
        s += len(g)
    return merge, labels


def merge_tokens(tokens: list, merge: list) -> list:
    """
    Merge tokens with an input merge

    References
    ----------
    Stack Overflow, https://stackoverflow.com/questions/43550219/merge-elements-in-list-based-on-given-indices
    """
    for t in merge[::-1]:
        merged = "".join(tokens[t[0] : t[1]])  # merging values within a range
        tokens[t[0] : t[1]] = [merged]  # slice replacement
    return tokens



import os
import os.path as op
import re
import joblib
import pandas as pd
from fuzzywuzzy import process
from spacy import displacy
from pythainlp import tokenize
from .utils import (
    preprocess,
    is_stopword,
    merge_tokens,
    merge_labels,
    get_digit,
    clean_location_text,
)
import warnings

warnings.filterwarnings("ignore", category=FutureWarning)


# read model from models path, define colors for output classes
MODULE_PATH = op.dirname(__file__)
CRF_MODEL = joblib.load(op.join(MODULE_PATH, "models", "model.joblib"))
ADDR_DF = pd.read_csv(
    op.join(MODULE_PATH, "data", "thai_address_data.csv"), dtype={"zipcode": str}
)
COLORS = {
    "NAME": "#fbd46d",
    "ADDR": "#ff847c",
    "LOC": "#87d4c5",
    "POST": "#def4f0",
    "PHONE": "#ffbffe",
    "EMAIL": "#91a6b8",
}
PROVINCES = list(ADDR_DF.province.unique()) + ["กรุงเทพ"]
DISTRICTS = list(ADDR_DF.district.unique())
SUBDISTRICTS = list(ADDR_DF.subdistrict.unique())
DISTRICTS_DICT = ADDR_DF.groupby("province")["district"].apply(list)
SUBDISTRICTS_DICT = ADDR_DF.groupby("province")["subdistrict"].apply(list)
DISTRICTS_POST_DICT = ADDR_DF.groupby("zipcode")["district"].apply(list)
SUBDISTRICTS_POST_DICT = ADDR_DF.groupby("zipcode")["subdistrict"].apply(list)


def extract_location(
    text: str, option="province", province=None, postal_code=None
) -> str:
    """
    Extract Thai province, district, or subdistrict
    from a given text by providing options

    Parameters
    ----------
    text: str, input Thai text of that contiains location
    option: str, an option to parse. This can be ``province``,
        ``district``, or ``subdistrict``
    province: str or None, if provided, we will only search
        for districts and subdistrcts within a given province
    
    Output
    ------
    location: str, output of location that best match with our
        primary text
    """
    # segment text
    text = text.replace("\n-", " ")
    text = text.replace("\n", " ")
    if option == "province":
        text = text.split("จ.")[-1].split("จังหวัด")[-1]
    elif option == "district":
        text = text.split("อ.")[-1].split("อำเภอ")[-1]
        text = text.split(" เขต")[-1]
    elif option == "subdistrict":
        text = text.split("ต.")[-1].split("อ.")[0].split("อำเภอ")[0]
        text = text.split(" แขวง")[-1].split(" เขต")[0]
    text = clean_location_text(text)

    location = ""
    if postal_code is not None and SUBDISTRICTS_POST_DICT.get(postal_code) is not None:
        options_map = {
            "province": PROVINCES,
            "district": DISTRICTS_POST_DICT.get(postal_code, DISTRICTS),
            "subdistrict": SUBDISTRICTS_POST_DICT.get(postal_code, SUBDISTRICTS),
        }
    elif province is not None:
        districts = []
        for d in DISTRICTS_DICT.get(province, DISTRICTS):
            if d != "พระนครศรีอยุธยา":
                districts.append(d.replace(province, ""))
            else:
                districts.append(d)
        options_map = {
            "province": PROVINCES,
            "district": districts,
            "subdistrict": SUBDISTRICTS_DICT.get(province, SUBDISTRICTS),
        }
    else:
        options_map = {
            "province": PROVINCES,
            "district": DISTRICTS,
            "subdistrict": SUBDISTRICTS,
        }
    options = options_map.get(option)
    try:
        locs = [l for l, _ in process.extract(text, options, limit=3)]
        locs.sort(key=len, reverse=False)  # sort from short to long string
        for loc in locs:
            if loc in text:
                location = loc
        if location == "" or location == "เมือง":
            location = [l for l, _ in process.extract(text, options, limit=3)][0]
    except:
        pass
    return location


def display_entities(tokens: list, labels: list):
    """
    Display tokens and labels

    References
    ----------
    Spacy, https://spacy.io/usage/visualizers
    """
    options = {"ents": list(COLORS.keys()), "colors": COLORS}

    ents = []
    text = ""
    s = 0
    for token, label in zip(tokens, labels):
        text += token
        if label != "O":
            ents.append({"start": s, "end": s + len(token), "label": label})
        s += len(token)

    text_display = {"text": text, "ents": ents, "title": None}
    displacy.render(
        text_display, style="ent", options=options, manual=True, jupyter=True
    )


def tokens_to_features(tokens: list, i: int) -> dict:
    """
    List of tokens to features for inputting to CRF suite
    """
    if len(tokens[i]) == 2:
        word, _ = tokens[i]  # unpack word and class
    else:
        word = tokens[i]

    # Features from current word
    features = {
        "bias": 1.0,
        "word.word": word,
        "word.isspace()": word.isspace(),
        "word.is_stopword()": is_stopword(word),
        "word.isdigit()": word.isdigit(),
    }
    if word.strip().isdigit() and len(word) == 5:
        features["word.islen5"] = True

    # Features from previous word
    if i > 0:
        prevword = tokens[i - 1][0]
        features.update(
            {
                "-1.word.prevword": prevword,
                "-1.word.isspace()": prevword.isspace(),
                "-1.word.is_stopword()": is_stopword(prevword),
                "-1.word.isdigit()": prevword.isdigit(),
            }
        )
    else:
        features["BOS"] = True  # Special "Beginning of Sequence" tag

    # Features from next word
    if i < len(tokens) - 1:
        nextword = tokens[i + 1][0]
        features.update(
            {
                "+1.word.nextword": nextword,
                "+1.word.isspace()": nextword.isspace(),
                "+1.word.is_stopword()": is_stopword(nextword),
                "+1.word.isdigit()": nextword.isdigit(),
            }
        )
    else:
        features["EOS"] = True  # Special "End of Sequence" tag

    return features


def parse(text: str, display: bool = False, tokenize_engine="deepcut") -> dict:
    """
    Parse a given address text and give a dictionary of
    parsed address out

    Parameters
    ----------
    text: str, input Thai address text to be parsed
    display: bool, if True, we will display parsed output
    tokenize_engine: str, pythainlp tokenization engines default is deepcut

    Output
    ------
    address: dict, parsed output 
    """
    text = preprocess(text)

    tokens = tokenize.word_tokenize(text, engine=tokenize_engine)
    features = [tokens_to_features(tokens, i) for i in range(len(tokens))]
    preds = CRF_MODEL.predict([features])[0]

    preds_ = list(zip(tokens, preds))
    name = "".join([token for token, c in preds_ if c == "NAME"]).strip()
    address = "".join([token for token, c in preds_ if c == "ADDR"]).strip()
    location = "".join([token for token, c in preds_ if c == "LOC"]).strip()

    postal_code = " ".join([token for token, c in preds_ if c == "POST"]).strip()
    postal_code = "".join([p for p in postal_code if p.isdigit()])

    if location != "":
        province = extract_location(location, option="province")
        if province == "กรุงเทพ":
            province = "กรุงเทพมหานคร"
        district = extract_location(
            location, option="district", province=province, postal_code=postal_code
        )
        subdistrict = extract_location(
            location, option="subdistrict", province=province, postal_code=postal_code
        )
    else:
        province = ""
        district = ""
        subdistrict = ""

    phone_number = " ".join(
        [get_digit(token) for token, c in preds_ if c == "PHONE" and len(token) > 1]
    ).strip()
    phone_number = "".join([p for p in phone_number if p.isdigit()])

    email = "".join([token for token, c in preds_ if c == "EMAIL"]).strip()
    if email == "":
        emails = re.findall(r"\b[\w.-]+?@\w+?\.\w+?\b", text)
        if len(emails) > 0:
            email = emails[0]

    # display parsed entities
    if display:
        merge, labels = merge_labels(preds)
        tokens = merge_tokens(tokens, merge)
        display_entities(tokens, labels)

    return {
        "text": text,
        "name": name,
        "address": address,
        "location": location,
        "subdistrict": subdistrict,
        "district": district,
        "province": province,
        "postal_code": postal_code,
        "phone_number": phone_number,
        "email": email,
    }



#! /usr/bin/env python
import os.path as op
from setuptools import setup, find_packages


def get_version():
    """
    Get the version without importing, so as not to invoke dependency
    requirements.
    """
    base, _ = op.split(op.realpath(__file__))
    file = op.join(base, "thaiaddress", "__init__.py")

    for line in open(file, "r"):
        if "__version__" in line:
            return line.split("=")[1].strip().strip("'").strip('"')


if __name__ == "__main__":
    setup(
        name="thaiaddress",
        version=get_version(),
        description="A Python parser for Thai address",
        long_description=open("README.md").read(),
        long_description_content_type='text/markdown',
        python_requires=">=3.6",
        url="https://github.com/425degree-developers/thaiaddress",
        download_url="https://github.com/425degree-developers/thaiaddress.git",
        author="Titipat Achakulvisut",
        author_email="my.titipat@gmail.com",
        license="Apache Software License 2.0 (c) 2020 Titipat Achakulvisut, 425 Degree Co., Bangkok, Thailand",
        install_requires=[
            "joblib",
            "deepcut",
            "spacy",
            "pythainlp",
            "sklearn_crfsuite",
            "numpy",
            "pytest",
            "pytest-cov",
            "scikit-learn",
            "jsonlines",
            "fuzzywuzzy",
            "pandas"
        ],
        packages=find_packages(),
        include_package_data=True,
        keywords=[
            "Parser",
            "Address",
            "Thai Address",
            "Thai Natural Language Processing",
            "Natural Language Processing",
        ],
        classifiers=[
            "Intended Audience :: Developers",
            "License :: OSI Approved :: Apache Software License",
            "Programming Language :: Python",
            "Topic :: Software Development",
            "Topic :: Scientific/Engineering",
            "Operating System :: Microsoft :: Windows",
            "Operating System :: Unix",
            "Operating System :: MacOS",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
        ],
        platforms="any",
        project_urls={
            "Source": "https://github.com/425degree-developers/thaiaddress",
            "Documentation": "https://github.com/425degree-developers/thaiaddress",
            "Bug Reports": "https://github.com/425degree-developers/thaiaddress/issues",
        },
    )
